import numpy
from scipy.optimize import differential_evolution

from catsim import irt, cat
from catsim.simulation import Estimator


class HillClimbingEstimator(Estimator):
    """Estimator that uses a hill-climbing algorithm to maximize the likelihood function

    :param precision: number of decimal points of precision
    :param verbose: verbosity level of the maximization method
    """

    def __str__(self):
        return 'Hill Climbing Estimator'

    def __init__(self, precision: int = 6, dodd: bool = False, verbose: bool = False):
        super().__init__()
        self._precision = precision
        self._verbose = verbose
        self._evaluations = 0
        self._calls = 0
        self._dodd = dodd

    @property
    def calls(self) -> float:
        """How many times the estimator has been called to maximize/minimize the log-likelihood function

        :returns: number of times the estimator has been called to maximize/minimize the log-likelihood function"""
        return self._calls

    @property
    def evaluations(self) -> float:
        """Total number of times the estimator has evaluated the log-likelihood function during its existence

        :returns: number of function evaluations"""
        return self._evaluations

    @property
    def avg_evaluations(self) -> float:
        """Average number of function evaluations for all tests the estimator has been used

        :returns: average number of function evaluations"""
        return self._evaluations / self._calls

    @property
    def dodd(self) -> bool:
        """Whether Dodd's method will be called by estimator in case the response vector
        is composed solely of right or wrong answers.

        :returns: boolean value indicating if Dodd's method will be used or not."""
        return self._dodd

    def estimate(self, index: int = None, items: numpy.ndarray = None, administered_items: list = None,
                 response_vector: list = None, est_theta: float = None) -> int:
        """Returns the theta value that minimizes the negative log-likelihood function, given the current state of the
         test for the given examinee.

        :param index: index of the current examinee in the simulator
        :param items: a matrix containing item parameters in the format that `catsim` understands (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :returns: the current :math:`\\hat\\theta`
        """
        if (index is None or self.simulator is None) and (
                                items is None and administered_items is None or response_vector is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if items is None and administered_items is None and response_vector is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            response_vector = self.simulator.response_vectors[index]
            est_theta = self.simulator.latest_estimations[index]

        self._calls += 1

        if len(set(response_vector)) == 1 and self._dodd:
            return cat.dodd(est_theta, self.simulator.items, response_vector[-1])

        if set(response_vector) == 1:
            return float('inf')
        elif set(response_vector) == 0:
            return float('-inf')

        if len(administered_items) > 0:
            lbound = min(items[administered_items][:, 1])
            ubound = max(items[administered_items][:, 1])
        else:
            lbound = min(items[:, 1])
            ubound = max(items[:, 1])

        best_theta = -float('inf')
        max_ll = -float('inf')

        self._evaluations = 0

        for _ in range(10):
            intervals = numpy.linspace(lbound, ubound, 10)
            if self._verbose:
                print(('Bounds: ' + str(lbound) + ' ' + str(ubound)))
                print(('Interval size: ' + str(intervals[1] - intervals[0])))

            for ii in intervals:
                self._evaluations += 1
                ll = irt.log_likelihood(ii, response_vector, items[administered_items])
                if ll > max_ll:
                    max_ll = ll

                    if self._verbose:
                        print(('Iteration: {0}, Theta: {1}, LL: {2}'.format(self._evaluations, ii, ll)))

                    if abs(best_theta - ii) < float('1e-' + str(self._precision)):
                        return ii

                    best_theta = ii

                else:
                    lbound = best_theta - (intervals[1] - intervals[0])
                    ubound = ii
                    break

        return best_theta


class DifferentialEvolutionEstimator(Estimator):
    """Estimator that uses :py:func:`scipy.optimize.differential_evolution` to minimize the negative log-likelihood function

    :param bounds: a tuple containing both lower and upper bounds for the differential
                   evolution algorithm search space. In theory, it is best if they
                   represent the minimum and maximum possible :math:`\\theta` values;
                   in practice, one could also use the smallest and largest difficulty
                   parameters in the item bank, in case no better bounds for
                   :math:`\\theta` exist.
    """

    def __str__(self):
        return 'Differential Evolution Estimator'

    def __init__(self, bounds: tuple):
        super(DifferentialEvolutionEstimator, self).__init__()
        self._lower_bound = min(bounds)
        self._upper_bound = max(bounds)
        self._evaluations = 0
        self._calls = 0

    @property
    def calls(self):
        """How many times the estimator has been called to maximize/minimize the log-likelihood function

        :returns: number of times the estimator has been called to maximize/minimize the log-likelihood function"""
        return self._calls

    @property
    def evaluations(self):
        """Total number of times the estimator has evaluated the log-likelihood function during its existence

        :returns: number of function evaluations"""
        return self._evaluations

    @property
    def avg_evaluations(self):
        """Average number of function evaluations for all tests the estimator has been used

        :returns: average number of function evaluations"""
        return self._evaluations / self._calls

    def estimate(self, index: int = None, items: numpy.ndarray = None, administered_items: list = None,
                 response_vector: list = None) -> int:
        """Uses :py:func:`scipy.optimize.differential_evolution` to return the theta value
        that minimizes the negative log-likelihood function, given the current state of the
        test for the given examinee.

        :param response_vector: a boolean list containing the examinee's answers to the administered items
        :param items: a matrix containing item parameters in the format that `catsim` understands (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param index: index of the current examinee in the simulator
        :returns: the current :math:`\\hat\\theta`
        """
        if (index is None or self.simulator is None) and (
                            items is None and administered_items is None or response_vector is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if items is None and administered_items is None and response_vector is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            response_vector = self.simulator.response_vectors[index]

        self._calls += 1

        res = differential_evolution(irt.negative_log_likelihood, bounds=[[self._lower_bound * 2, self._upper_bound * 2]],
                                     args=(response_vector, items[administered_items]))

        self._evaluations = res.nfev

        return res.x[0]



import numpy

from catsim.simulation import Initializer


class RandomInitializer(Initializer):
    """Randomly initializes the first estimate of an examinee's proficiency

    :param dist_type: either `uniform` or `normal`
    :param dist_params: a tuple containing minimum and maximum values for the
                        uniform distribution (in no particular order) or the average
                        and standard deviation values for the normal distribution
                        (in this particular order)."""

    def __str__(self):
        return 'Random Initializer'

    def __init__(self, dist_type: str = 'uniform', dist_params: tuple = (-5, 5)):
        super(RandomInitializer, self).__init__()

        available_distributions = ['uniform', 'normal']
        if dist_type not in available_distributions:
            raise ValueError('{0} not in available distributions {1}'.format(dist_type, available_distributions))

        self._dist_type = dist_type
        self._dist_params = dist_params

    def initialize(self, index: int) -> float:
        """Generates a value using the chosen distribution and parameters

        :param index: the index of the current examinee. This parameter is not used by this method.
        :returns: a proficiency value generated from the chosen distribution using the passed parameters"""
        if self._dist_type == 'uniform':
            return numpy.random.uniform(min(self._dist_params), max(self._dist_params))
        elif self._dist_type == 'normal':
            return numpy.random.normal(self._dist_params[0], self._dist_params[1])


class FixedPointInitializer(Initializer):
    """Initializes every proficiency at the same point."""

    def __str__(self):
        return 'Fixed Point Initializer'

    def __init__(self, start: float):
        """
        :param start: the starting point for every examinee
        """
        super(FixedPointInitializer, self).__init__()
        self._start = start

    def initialize(self, index: int) -> float:
        """Returns the same proficiency value that was passed to the constructor of the initializer

        :param index: the index of the current examinee. This parameter is not used by this method.
        :returns: the same proficiency value that was passed to the constructor of the initializer"""
        return self._start



import numpy
from scipy.integrate import quad

from catsim import irt
from catsim.simulation import Selector


class MaxInfoSelector(Selector):
    """Selector that returns the first non-administered item with maximum information, given an estimated theta"""

    def __init__(self):
        super().__init__()

    def __str__(self):
        return 'Maximum Information Selector'

    def select(self, index: int = None, items: numpy.ndarray = None, administered_items: list = None,
               est_theta: float = None) -> int:
        """Returns the index of the next item to be administered.

        :param items: a matrix containing item parameters in the format that `catsim` understands (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None) and (
                            items is None or administered_items is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if items is None and administered_items is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            est_theta = self.simulator.latest_estimations[index]

        valid_indexes = [x for x in range(items.shape[0]) if x not in administered_items]
        inf_values = [irt.inf(est_theta, i[0], i[1], i[2], i[3]) for i in items[valid_indexes]]
        valid_indexes = [item_index for (inf_value, item_index) in
                         sorted(zip(inf_values, valid_indexes), key=lambda pair: pair[0], reverse=True)]

        return valid_indexes[0]


class LinearSelector(Selector):
    """Selector that returns item indexes in a linear order, simulating a standard
    (non-adaptive) test.

    :param indexes: the indexes of the items that will be returned in order"""

    def __str__(self):
        return 'Linear Selector'

    def __init__(self, indexes: list):
        super().__init__()
        self._indexes = indexes
        self._current = 0

    @property
    def indexes(self):
        return self._indexes

    @property
    def current(self):
        return self._current

    def select(self, index: int = None, administered_items: list = None) -> int:
        """Returns the index of the next item to be administered.

        :param administered_items: a list containing the indexes of items that were already administered
        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None) and (administered_items is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if administered_items is None:
            administered_items = self.simulator.administered_items[index]

        if set(self._indexes).issubset(set(administered_items)):
            raise ValueError(
                'A new index was asked for, but there are no more item indexes to present.\nCurrent item:\t\t\t{0}\nItems to be administered:\t{1} (size: {2})\nAdministered items:\t\t{3} (size: {4})'.format(
                    self._current, sorted(self._indexes), len(self._indexes), sorted(administered_items),
                    len(administered_items)))

        selected_item = [x for x in self._indexes if x not in administered_items][0]

        return selected_item


class RandomSelector(Selector):
    """Selector that randomly selects items for application.

    :param replace: whether to select an item that has already been selected before for this examinee."""

    def __str__(self):
        return 'Random Selector'

    def __init__(self, replace: bool = False):
        super().__init__()
        self._replace = replace

    def select(self, index: int = None, items: numpy.ndarray = None, administered_items: list = None) -> int:
        """Returns the index of the next item to be administered.

        :param items: a matrix containing item parameters in the format that `catsim` understands (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None) and (items is None or administered_items is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if items is None and administered_items is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]

        if len(administered_items) >= items.shape[0] and not self._replace:
            raise ValueError('A new item was asked for, but there are no more items to present.')

        if self._replace:
            return numpy.random.choice(items.shape[0])
        else:
            return numpy.random.choice(list(set(range(items.shape[0])) - set(administered_items)))


class ClusterSelector(Selector):
    """Cluster-based Item Selection Method.

        .. [Men15] Meneghetti, D. R. (2015). Metolodogia de seleção de itens em testes
           adaptativos informatizados baseada em agrupamento por similaridade (Mestrado).
           Centro Universitário da FEI. Retrieved from
           https://www.researchgate.net/publication/283944553_Metodologia_de_selecao_de_itens_em_Testes_Adaptativos_Informatizados_baseada_em_Agrupamento_por_Similaridade

    :param clusters: a list containing item cluster memberships
    :param r_max: maximum exposure rate for items
    :param method: one of the available methods for cluster selection. Given
                   the estimated theta value at each step:

                       ``item_info``: selects the cluster which has the item
                       with maximum information;

                       ``cluster_info``: selects the cluster whose items sum of
                       information is maximum;

                       ``weighted_info``: selects the cluster whose weighted
                       sum of information is maximum. The weighted equals the
                       number of items in the cluster;

    :param r_control: if `passive` and all items :math:`i` in the selected
                      cluster have :math:`r_i > r^{max}`, applies the item with
                      maximum information; if `aggressive`, applies the item
                      with smallest :math:`r` value.
    """

    def __str__(self):
        return 'Cluster Selector'

    def __init__(self, clusters: list, method: str = 'item_info', r_max: float = 1, r_control: str = 'passive'):
        super().__init__()
        available_methods = ['item_info', 'cluster_info', 'weighted_info']
        if method not in available_methods:
            raise ValueError(
                '{0} is not a valid cluster selection method; choose one from {1}'.format(method, available_methods))
        available_rcontrol = ['passive', 'aggressive']
        if r_control not in available_rcontrol:
            raise ValueError('{0} is not a valid item exposure control method; choose one from {1}'.format(r_control,
                                                                                                           available_rcontrol))

        self._clusters = clusters
        self._method = method
        self._r_max = r_max
        self._r_control = r_control

    def select(self, index: int = None, items: numpy.ndarray = None, administered_items: list = None,
               est_theta: float = None) -> int:
        """Returns the index of the next item to be administered.

        :param items: a matrix containing item parameters in the format that `catsim` understands (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None) and (
                            items is None or administered_items is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if items is None and administered_items is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            est_theta = self.simulator.latest_estimations[index]

        selected_cluster = None
        # this part of the code selects the cluster from which the item at
        # the current point of the test will be chosen
        if self._method == 'item_info':
            infos = [irt.inf(est_theta, i[0], i[1], i[2], i[3]) for i in items]

            while selected_cluster is None:
                # find item with maximum information
                max_info_item = infos.index(max(infos))

                # gets the indexes of all the items in the same cluster
                # as the current selected item that have not been
                # administered
                valid_indexes = [i for i, x in enumerate(self._clusters) if x == self._clusters[max_info_item]]

                # if all items in the same cluster as the selected have been used,
                # get the next item with maximum information
                if set(valid_indexes).issubset(set(administered_items)):
                    infos[max_info_item] = float('-inf')
                else:
                    selected_cluster = self._clusters[max_info_item]

        elif self._method in ['cluster_info', 'weighted_info']:
            # calculates the cluster information, depending on the method
            # selected
            if self._method == 'cluster_info':
                cluster_infos = ClusterSelector.sum_cluster_infos(est_theta, items, self._clusters)
            elif self._method == 'weighted_info':
                cluster_infos = ClusterSelector.weighted_cluster_infos(est_theta, items, self._clusters)

            # sorts clusters descending by their information values
            # this type of sorting was seem on
            # http://stackoverflow.com/a/6618543
            sorted_clusters = numpy.array([cluster for (inf_value, cluster) in
                                           sorted(zip(cluster_infos, set(self._clusters)), key=lambda pair: pair[0],
                                                  reverse=True)], dtype=float)

            # walks through the sorted clusters in order
            for i in range(len(sorted_clusters)):
                valid_indexes = numpy.nonzero(items[:, 4] == sorted_clusters[i])[0]

                # checks if at least one item from this cluster has not
                # been administered to this examinee yet
                if set(valid_indexes).intersection(administered_items) != set(valid_indexes):
                    selected_cluster = sorted_clusters[i]
                    break
                    # the for loop ends with the cluster that has a) the maximum
                    # information possible and b) at least one item that has not
                    # yet been administered

        # in this part, an item is chosen from the cluster that was
        # selected above

        # gets the indexes and information values from the items in the
        # selected cluster that have not been administered
        valid_indexes = numpy.array(
            list(set(numpy.nonzero(self._clusters == selected_cluster)[0]) - set(administered_items)))

        # gets the indexes and information values from the items in the
        # selected cluster with r < rmax that have not been
        # administered
        valid_indexes_low_r = numpy.array(list(
            set(numpy.nonzero((self._clusters == selected_cluster) & (items[:, 4] < self._r_max))[0]) - set(
                administered_items)))

        if len(valid_indexes_low_r) > 0:
            # sort both items and their indexes by their information
            # value
            inf_values = [irt.inf(est_theta, i[0], i[1], i[2], i[3]) for i in items[valid_indexes_low_r]]
            valid_indexes_low_r = [item_index for (inf_value, item_index) in
                                   sorted(zip(inf_values, valid_indexes_low_r), key=lambda pair: pair[0], reverse=True)]

            selected_item = valid_indexes_low_r[0]

        # if all items in the selected cluster have exceed their r values,
        # select the one with smallest r, regardless of information
        else:
            if self._r_control == 'passive':
                inf_values = [irt.inf(est_theta, i[0], i[1], i[2], i[3]) for i in items[valid_indexes]]
                valid_indexes = [item_index for (inf_value, item_index) in
                                 sorted(zip(inf_values, valid_indexes), key=lambda pair: pair[0], reverse=True)]
            elif self._r_control == 'aggressive':
                valid_indexes = [item_index for (r, item_index) in
                                 sorted(zip(items[valid_indexes, 4], valid_indexes), key=lambda pair: pair[0])]

            selected_item = valid_indexes[0]

        return selected_item

    @staticmethod
    def sum_cluster_infos(theta: float, items: numpy.ndarray, clusters: list) -> list:
        """Returns the sum of item informations, separated by cluster

        :param theta: an examinee's :math:`\\theta` value
        :param items: a matrix containing item parameters in the format that `catsim` understands (see: :py:func:`catsim.cat.generate_item_bank`)
        :param clusters: a list containing item cluster memberships, represented by integers
        :returns: list containing the sum of item information values for each cluster"""
        cluster_infos = numpy.zeros((len(set(clusters))))

        for cluster in set(clusters):
            cluster_indexes = numpy.nonzero(clusters == cluster)[0]

            for item in items[cluster_indexes]:
                cluster_infos[cluster] += irt.inf(theta, item[0], item[1], item[2], item[3])

        return cluster_infos

    @staticmethod
    def weighted_cluster_infos(theta: float, items: numpy.ndarray, clusters: list):
        """Returns the weighted sum of item informations, separated by cluster.
        The weight is the number of items in each cluster.

        :param theta: an examinee's :math:`\\theta` value
        :param items: a matrix containing item parameters in the format that `catsim` understands (see: :py:func:`catsim.cat.generate_item_bank`)
        :param clusters: a list containing item cluster memberships, represented by integers
        :returns: list containing the sum of item information values for each cluster,
                  divided by the number of items in each cluster"""
        cluster_infos = ClusterSelector.sum_cluster_infos(theta, items, clusters)
        count = numpy.bincount(clusters)

        for i in range(len(cluster_infos)):
            cluster_infos[i] /= count[i]

        return cluster_infos

    @staticmethod
    def sum_cluster_params(items: numpy.ndarray, c: list):
        """Returns the sum of item parameter values for each cluster cluster

        :param items: a matrix containing item parameters in the format that `catsim` understands (see: :py:func:`catsim.cat.generate_item_bank`)
        :param c: a list containing clustering memeberships.
        :returns: a matrix containing the sum of each parameter by cluster. Lines are clusters, columns are parameters.
        """
        averages = numpy.zeros((numpy.max(c) + 1, 4))

        for i in numpy.arange(0, numpy.size(c)):
            if c[i] == -1:
                continue
            averages[c[i], 0] += items[i, 0]
            averages[c[i], 1] += items[i, 1]
            averages[c[i], 2] += items[i, 2]
            averages[c[i], 3] += items[i, 3]

        return averages

    @staticmethod
    def avg_cluster_params(items: numpy.ndarray, c: list):
        """Returns the average values of item parameters by cluster

        :param items:
        :param c: a list containing clustering memeberships.
        :returns: a matrix containing the average values of each parameter by cluster.
                  Lines are clusters, columns are parameters."""
        averages = ClusterSelector.sum_cluster_params(items, c)

        occurrences = numpy.bincount(numpy.delete(c, numpy.where(c == -1)).astype(numpy.int64))

        for counter, i in enumerate(occurrences):
            averages[counter, 0] /= i
            averages[counter, 1] /= i
            averages[counter, 2] /= i
            averages[counter, 3] /= i

        return averages


class AStratifiedSelector(Selector):
    """Implementation of the :math:`\\alpha`-stratified selector proposed by
    [Chang99]_, in which the item bank is sorted in ascending order according to the
    items discrimination parameter and then separated into :math:`K` strata
    (:math:`K` being the test size), each stratum containing gradually higher
    average discrimination. The :math:`\\alpha`-stratified selector then selects the
    first non-administered item from stratum :math:`k`, in which :math:`k`
    represents the position in the test of the current item the examinee is being
    presented.

    .. image:: ../docs/alpha-strat.*

    :param test_size: the number of items the test contains. The selector uses this parameter
    to create the correct number of strata.
    """

    def __str__(self):
        return 'a-Stratified Selector'

    def __init__(self, test_size):
        super().__init__()
        self._organized_items = None
        self._test_size = test_size

    @property
    def test_size(self):
        return self._test_size

    def preprocess(self):
        # sort item indexes by their discrimination value
        self._organized_items = self.simulator.items[:, 0].argsort()

    def select(self, index: int = None, administered_items: list = None) -> int:
        """Returns the index of the next item to be administered.

        :param administered_items: a list containing the indexes of items that were already administered
        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None) and (administered_items is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if administered_items is None:
            administered_items = self.simulator.administered_items[index]

        # select the item in the correct layer, according to the point in the test the examinee is
        pointer = len(administered_items) * self._test_size

        # if the selected item has already been administered, select the next one
        while self._organized_items[pointer] in administered_items:
            pointer += 1

        return self._organized_items[pointer]


class AStratifiedBBlockingSelector(Selector):
    """Implementation of the :math:`\\alpha`-stratified selector with :math:`b`
    blocking proposed by [Chang2001]_, in which the item bank is sorted in ascending
    order according to the items difficulty parameter and then separated into
    :math:`M` strata, each stratum containing gradually higher average difficulty.

    Each of the :math:`M` strata is then again separated into :math:`K`
    sub-strata (:math:`k` being the test size), according to their
    discrimination. The final item bank is then ordered such that the first
    sub-strata of each strata forms the first strata of the new ordered item
    bank, and so on. This method tries to balance the distribution of both
    parameters between all strata, after perceiving that they are correlated.

    .. image:: ../docs/b-blocking.*

    :param test_size: the number of items the test contains. The selector uses this parameter to
    create the correct number of strata.
    """

    def __str__(self):
        return 'a-Stratified b-Blocking Selector'

    def __init__(self, test_size):
        super().__init__()
        self._organized_items = None
        self._test_size = test_size

    @property
    def test_size(self):
        return self._test_size

    def preprocess(self):
        # sort item indexes by their difficulty, then their discrimination value
        self._organized_items = numpy.lexsort((self.simulator.items[:, 0], self.simulator.items[:, 1]))

    def select(self, index: int = None, administered_items: list = None) -> int:
        """Returns the index of the next item to be administered.

        :param administered_items: a list containing the indexes of items that were already administered
        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None) and (administered_items is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if administered_items is None:
            administered_items = self.simulator.administered_items[index]

        # select the item in the correct layer, according to the point in the test the examinee is
        selected_item = len(administered_items)

        # if the selected item has already been administered, select the next one
        while self._organized_items[selected_item] in administered_items:
            selected_item += self._test_size

        return self._organized_items[selected_item]


class MaxInfoStratificationSelector(Selector):
    """Implementation of the maximum information stratification (MIS) selector
    proposed by [Bar06]_, in which the item bank is sorted in ascending order
    according to the items maximum information and then separated into :math:`K`
    strata (:math:`K` being the test size), each stratum containing items with
    gradually higher maximum information. The MIS selector then selects the first
    non-administered item from stratum :math:`k`, in which :math:`k` represents the
    position in the test of the current item the examinee is being presented.

    .. image:: ../docs/mis.*

    This method claims to work better than the :math:`a`-stratified method by
    [Chang99]_ for the three-parameter logistic model of IRT, since item difficulty
    and maximum information are not positioned in the same place in the proficiency
    scale in 3PL.

    :param test_size: the number of items the test contains. The selector uses this parameter to
    create the correct number of strata.
    """

    def __str__(self):
        return 'Maximum Information Stratification Selector'

    def __init__(self, test_size):
        super().__init__()
        self._organized_items = None
        self._test_size = test_size

    @property
    def test_size(self):
        return self._test_size

    def preprocess(self):
        # sort item indexes by their maximum information value
        self._organized_items = numpy.array(
            [irt.inf(irt.max_info(item[0], item[1], item[2], item[3]), item[0], item[1], item[2], item[3]) for item in
             self.simulator.items]).argsort()

    def select(self, index: int = None, administered_items: list = None) -> int:
        """Returns the index of the next item to be administered.

        :param administered_items: a list containing the indexes of items that were already administered
        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None) and (administered_items is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if administered_items is None:
            administered_items = self.simulator.administered_items[index]

        # select the item in the correct layer, according to the point in the test the examinee is
        pointer = len(administered_items) * self._test_size

        # if the selected item has already been administered, select the next one
        while self._organized_items[pointer] in administered_items:
            pointer += 1

        return self._organized_items[pointer]


class MaxInfoBBlockingSelector(Selector):
    """Implementation of the maximum information stratification with :math:`b`
    blocking (MIS-B) selector proposed by [Bar06]_, in which the item bank is sorted
    in ascending order according to the items difficulty parameter and then
    separated into :math:`M` strata, each stratum containing gradually higher
    average difficulty.

    Each of the :math:`M` strata is then again separated into :math:`K`
    sub-strata (:math:`k` being the test size), according to the items maximum
    information. The final item bank is then ordered such that the first
    sub-strata of each strata forms the first strata of the new ordered item
    bank, and so on. This method tries to balance the distribution of both
    parameters between all strata and works better than the :math:`a`-stratified
    with :math:`b` blocking method by [Chang2001]_ for the three-parameter
    logistic model of IRT, since item difficulty and maximum information are not
    positioned in the same place in the proficiency scale in 3PL. This may also
    apply, although not mentioned by the authors, for the 4PL.

    .. image:: ../docs/mis-b.*

    :param test_size: the number of items the test contains. The selector uses this parameter to
    create the correct number of strata.
    """

    def __str__(self):
        return 'Maximum Information Stratification with b-Blocking Selector'

    def __init__(self, test_size):
        super().__init__()
        self._organized_items = None
        self._test_size = test_size

    @property
    def test_size(self):
        return self._test_size

    def preprocess(self):
        # sort item indexes by their theta_max values, then by their maximum information
        self._organized_items = numpy.lexsort(([irt.inf(irt.max_info(item[0], item[1], item[2], item[3]), item[0],
                                                        item[1], item[2], item[3]) for item in self.simulator.items],
                                               [irt.max_info(item[0], item[1], item[2], item[3]) for item in
                                                self.simulator.items]))

    def select(self, index: int = None, administered_items: list = None) -> int:
        """Returns the index of the next item to be administered.

        :param administered_items: a list containing the indexes of items that were already administered
        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None) and (administered_items is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if administered_items is None:
            administered_items = self.simulator.administered_items[index]

        # select the item in the correct layer, according to the point in the test the examinee is
        pointer = len(administered_items)

        # if the selected item has already been administered, select the next one
        while self._organized_items[pointer] in administered_items:
            pointer += self._test_size

        return self._organized_items[pointer]


class The54321Selector(Selector):
    """Implementation of the 5-4-3-2-1 selector proposed by [McBride83]_, in which,
    at each step :math:`k` of a test of size :math:`K`, an item is chosen from a bin
    containing the :math:`K-k` most informative items in the bank, given the current
    :math:`\\hat\\theta`. As the test progresses, the bin gets smaller and more
    informative items have a higher probability of being chosen by the end of the
    test, when the estimation of ':math:`\\hat\\theta` is more precise. The
    5-4-3-2-1 selector can be viewed as a specialization of the
    :py:class:`catsim.selection.RandomesqueSelector`, in which the bin size of most
    informative items gets smaller as the test progresses.

    :param test_size: the number of items the test contains. The selector uses
                      this parameter to set the bin size"""

    def __str__(self):
        return '5-4-3-2-1 Selector'

    def __init__(self, test_size):
        super().__init__()
        self._test_size = test_size

    @property
    def test_size(self):
        return self._test_size

    def select(self, index: int = None, items: numpy.ndarray = None, administered_items: list = None,
               est_theta: float = None) -> int:
        """Returns the index of the next item to be administered.

        :param items: a matrix containing item parameters in the format that `catsim` understands (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None) and (
                            items is None or administered_items is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if items is None and administered_items is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            est_theta = self.simulator.latest_estimations[index]

        # sort item indexes by their information value and remove indexes of administered items
        organized_items = [x for x in numpy.array(
            [irt.inf(est_theta, item[0], item[1], item[2], item[3]) for item in items]).argsort() if
                           x not in administered_items]

        bin_size = self._test_size - len(administered_items)

        if len(organized_items) == 0:
            raise ValueError('There are no more items to apply.')

        return numpy.random.choice(organized_items[0:bin_size])


class RandomesqueSelector(Selector):
    """Implementation of the randomesque selector proposed by [Kingsbury89]_, in which,
    at every step of the test, an item is randomly chosen from the :math:`n` most informative
    items in the item bank, :math:`n` being a predefined value (originally 5, but user-defined
    in this implementation)

    :param bin_size: the number of most informative items to be taken into consideration when
    randomly selecting one of them.
    """

    def __str__(self):
        return 'Randomesque Selector'

    def __init__(self, bin_size):
        super().__init__()
        self._bin_size = bin_size

    @property
    def bin_size(self):
        return self._bin_size

    def select(self, index: int = None, items: numpy.ndarray = None, administered_items: list = None,
               est_theta: float = None) -> int:
        """Returns the index of the next item to be administered.

        :param items: a matrix containing item parameters in the format that `catsim` understands (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None) and (
                            items is None or administered_items is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if items is None and administered_items is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            est_theta = self.simulator.latest_estimations[index]

        # sort item indexes by their information value and remove indexes of administered items
        organized_items = [x for x in numpy.array(
            [irt.inf(est_theta, item[0], item[1], item[2], item[3]) for item in items]).argsort() if
                           x not in administered_items]

        if len(organized_items) == 0:
            raise ValueError('There are no more items to apply.')

        return numpy.random.choice(list(organized_items)[:self._bin_size])


class IntervalIntegrationSelector(Selector):
    """Implementation of an interval integration selector in which, at every step of
    the test, the item that maximizes the information function integral at a
    predetermined ``interval`` :math:`\\delta` above and below the current
    :math:`\\hat\\theta` is chosen.

    .. math:: argmax_{i \\in I} \\int_{\\hat\\theta - \\delta}^{\\hat\\theta - \\delta}I_i(\\hat\\theta)

    :param interval: the interval of the integral. If no interval is passed, the
                     integral is calculated from :math:`[-\\infty, \\infty]`.
    """

    def __str__(self):
        return 'Interval Integration Selector'

    def __init__(self, interval: float = None):
        super().__init__()
        self._interval = interval if interval is not None else numpy.inf

    @property
    def interval(self):
        return self._interval

    def select(self, index: int = None, items: numpy.ndarray = None, administered_items: list = None,
               est_theta: float = None) -> int:
        """Returns the index of the next item to be administered.

        :param items: a matrix containing item parameters in the format that `catsim` understands (see: :py:func:`catsim.cat.generate_item_bank`)
        :param administered_items: a list containing the indexes of items that were already administered
        :param est_theta: a float containing the current estimated proficiency
        :param index: the index of the current examinee in the simulator.
        :returns: index of the next item to be applied.
        """
        if (index is None or self.simulator is None) and (
                            items is None or administered_items is None or est_theta is None):
            raise ValueError(
                'Either pass an index for the simulator, or the item bank, administered_items and est_theta to select the next item independently.')

        if items is None and administered_items is None and est_theta is None:
            items = self.simulator.items
            administered_items = self.simulator.administered_items[index]
            est_theta = self.simulator.latest_estimations[index]

        # sort item indexes by the integral of the information function and remove indexes of administered items
        organized_items = [x for x in numpy.array([quad(irt.inf, est_theta - self._interval, est_theta + self._interval,
                                                        args=(item[0], item[1], item[2], item[3]))[0] for item in
                                                   items]).argsort() if x not in administered_items]

        if len(organized_items) == 0:
            raise ValueError('There are no more items to apply.')

        return list(organized_items)[0]



"""Functions used specifically during the application/simulation of computerized adaptive tests."""

import operator

import numpy

from catsim import irt


def dodd(theta: float, items: numpy.ndarray, correct: bool) -> float:
    """Method proposed by [Dod90]_ for the reestimation of :math:`\\hat{\\theta}`
    when the response vector is composed entirely of 1s or 0s.

    .. math::

        \\hat{\\theta}_{t+1} = \\left\\lbrace \\begin{array}{ll}
        \\hat{\\theta}_t+\\frac{b_{max}-\\hat{\\theta_t}}{2} & \\text{if } X_t = 1 \\\\
        \\hat{\\theta}_t-\\frac{\\hat{\\theta}_t-b_{min}}{2} & \\text{if }  X_t = 0
        \\end{array} \\right\\rbrace

    :param theta: the initial proficiency level
    :param items: a numpy array containing the parameters of the items in the
                  database. This is necessary to capture the maximum and minimum
                  difficulty levels necessary for the method.
    :param correct: a boolean value informing whether or not the examinee
                    correctly answered the current item.
    :returns: a new estimation for :math:`\\theta`
    """
    b = items[:, 1]

    return theta + ((max(b) - theta) / 2) if correct else theta - ((theta - min(b)) / 2)


def bias(actual: list, predicted: list):
    """Calculates the test bias, an evaluation criterion for computerized adaptive test methodolgies [Chang2001]_.
    The value is calculated by:

    .. math:: Bias = \\frac{\\sum_{i=1}^{N} (\\hat{\\theta}_i - \\theta_{i})}{N}

    where :math:`\\hat{\\theta}_i` is examinee :math:`i` estimated proficiency and
    :math:`\\hat{\\theta}_i` is examinee :math:`i` actual proficiency.

    :param actual: a list or 1-D numpy array containing the true proficiency values
    :param predicted: a list or 1-D numpy array containing the estimated proficiency values
    :returns: the bias between the predicted values and actual values.
    """
    if len(actual) != len(predicted):
        raise ValueError('actual and predicted vectors need to be the same size')
    return numpy.mean(list(map(operator.sub, predicted, actual)))


def mse(actual: list, predicted: list):
    """Mean squared error, a value used when measuring the precision
    with which a computerized adaptive test estimates examinees proficiencies [Chang2001]_.
    The value is calculated by:

    .. math:: MSE = \\frac{\\sum_{i=1}^{N} (\\hat{\\theta}_i - \\theta_{i})^2}{N}

    where :math:`\\hat{\\theta}_i` is examinee :math:`i` estimated proficiency and
    :math:`\\hat{\\theta}_i` is examinee :math:`i` actual proficiency.

    :param actual: a list or 1-D numpy array containing the true proficiency values
    :param predicted: a list or 1-D numpy array containing the estimated proficiency values
    :returns: the mean squared error between the predicted values and actual values.
    """
    if len(actual) != len(predicted):
        raise ValueError('actual and predicted vectors need to be the same size')
    return numpy.mean([x * x for x in list(map(operator.sub, predicted, actual))])


def rmse(actual: list, predicted: list):
    """Root mean squared error, a common value used when measuring the precision
    with which a computerized adaptive test estimates examinees proficiencies [Bar10]_.
    The value is calculated by:

    .. math:: RMSE = \\sqrt{\\frac{\\sum_{i=1}^{N} (\\hat{\\theta}_i - \\theta_{i})^2}{N}}

    where :math:`\\hat{\\theta}_i` is examinee :math:`i` estimated proficiency and
    :math:`\\hat{\\theta}_i` is examinee :math:`i` actual proficiency.

    :param actual: a list or 1-D numpy array containing the true proficiency values
    :param predicted: a list or 1-D numpy array containing the estimated proficiency values
    :returns: the root mean squared error between the predicted values and actual values.
    """
    if len(actual) != len(predicted):
        raise ValueError('actual and predicted vectors need to be the same size')
    return numpy.sqrt(mse(actual, predicted))


def overlap_rate(items: numpy.ndarray, test_size: int) -> float:
    """Test overlap rate, an average measure of how much of the test two examinees take is equal [Bar10]_. It is given by:

    .. math:: T=\\frac{N}{Q}S_{r}^2 + \\frac{Q}{N}

    If, for example :math:`T = 0.5`, it means that the tests of two random examinees have 50% of equal items.

    :param items: a matrix containing, in the 4th column, the number of
                  times each item was used in the tests.
    :param test_size: an integer informing the number of items in a test.
    :returns: test overlap rate.
    """

    bank_size = items.shape[0]
    var_r = numpy.var(items[:, 3])

    t = (bank_size / test_size) * var_r + (test_size / bank_size)

    return t


def generate_item_bank(n: int, itemtype: str = '4PL', corr: float = 0.5):
    """Generate a synthetic item bank whose parameters approximately follow
    real-world parameters, as proposed by [Bar10]_.

    Item parameters are extracted from the following probability distributions:

    * discrimination: :math:`N(1.2, 0.25)`

    * difficulty: :math:`N(0,  1)`

    * pseudo-guessing: :math:`N(0.25, 0.02)`

    * upper asymptote: :math:`U(0.94, 1)`

    :param n: how many items are to be generated
    :param itemtype: either ``1PL``, ``2PL``, ``3PL`` or ``4PL`` for the one-, two-,
                     three- or four-parameter logistic model
    :param corr: the correlation between item discrimination and difficulty. If
                 ``itemtype == '1PL'``, it is ignored.
    :return: an ``n x 4`` numerical matrix containing item parameters
    :rtype: numpy.ndarray

    >>> generate_item_bank(5, '1PL')
    >>> generate_item_bank(5, '2PL')
    >>> generate_item_bank(5, '3PL')
    >>> generate_item_bank(5, '4PL')
    >>> generate_item_bank(5, '4PL', corr=0)
    """

    valid_itemtypes = ['1PL', '2PL', '3PL', '4PL']

    if itemtype not in valid_itemtypes:
        raise ValueError('Item type not in ' + str(valid_itemtypes))

    means = [0, 1.2]
    stds = [1, 0.25]
    covs = [[stds[0] ** 2, stds[0] * stds[1] * corr], [stds[0] * stds[1] * corr, stds[1] ** 2]]

    b, a = numpy.random.multivariate_normal(means, covs, n).T

    # if by chance there is some discrimination value below zero
    # this makes the problem go away
    if any(disc < 0 for disc in a):
        min_disc = min(a)
        a = [disc + abs(min_disc) for disc in a]

    if itemtype not in ['2PL', '3PL', '4PL']:
        a = numpy.ones(n)

    if itemtype in ['3PL', '4PL']:
        c = numpy.random.normal(.25, .02, n)
    else:
        c = numpy.zeros(n)

    if itemtype == '4PL':
        d = numpy.random.uniform(.94, 1, n)
    else:
        d = numpy.ones(n)

    return irt.normalize_item_bank(numpy.array([a, b, c, d]).T)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



"""Module with functions for plotting IRT-related results."""

import os

import matplotlib.pyplot as plt
import numpy
from mpl_toolkits.mplot3d import Axes3D

from catsim import irt
from catsim.simulation import Simulator


def __column(matrix, i):
    """Returns columns from a bidimensional Python list (a list of lists)"""
    return [row[i] for row in matrix]


def item_curve(a: float = 1, b: float = 0, c: float = 0, d: float = 1, title: str = None, ptype: str = 'icc',
               max_info=True, filepath: str = None, show: bool = True):
    """Plots 'Item Response Theory'-related item plots

    .. plot::

        from catsim.cat import generate_item_bank
        from catsim import plot
        item = generate_item_bank(1)[0]
        plot.item_curve(item[0], item[1], item[2], item[3], ptype='icc')
        plot.item_curve(item[0], item[1], item[2], item[3], ptype='iic')
        plot.item_curve(item[0], item[1], item[2], item[3], ptype='both')

    When both curves are plotted in the same figure, the figure has no grid,
    since each curve has a different scale.

    :param a: item discrimination parameter
    :param b: item difficulty parameter
    :param c: item pseudo-guessing parameter
    :param d: item upper asymptote
    :param title: plot title
    :param ptype: 'icc' for the item characteristic curve, 'iic' for the item
                  information curve or 'both' for both curves in the same plot
    :param max_info: whether the point of maximum information should be shown in the plot
    :param filepath: saves the plot in the given path
    :param show: whether the generated plot is to be shown
    """
    available_types = ['icc', 'iic', 'both']

    if ptype not in available_types:
        raise ValueError('\'{0}\' not in available plot types: {1}'.format(ptype, available_types))

    thetas = numpy.arange(b - 4, b + 4, .1, 'double')
    p_thetas = []
    i_thetas = []
    for theta in thetas:
        p_thetas.append(irt.icc(theta, a, b, c, d))
        i_thetas.append(irt.inf(theta, a, b, c, d))

    if ptype in ['icc', 'iic']:
        plt.figure()

        if title is not None:
            plt.title(title, size=18)

        plt.annotate(
            '$a = ' + format(a) + '$\n$b = ' + format(b) + '$\n$c = ' + format(c) + '$\n$d = ' + format(d) + '$',
            bbox=dict(facecolor='white', alpha=1), xy=(.75, .05), xycoords='axes fraction')
        plt.xlabel(r'$\theta$')
        plt.grid()

        if ptype == 'icc':
            plt.ylabel(r'$P(\theta)$')
            plt.plot(thetas, p_thetas, label=r'$P(\theta)$')

        elif ptype == 'iic':
            plt.ylabel(r'$I(\theta)$')
            plt.plot(thetas, i_thetas, label=r'$I(\theta)$')
            if max_info:
                aux = irt.max_info(a, b, c, d)
                plt.plot(aux, irt.inf(aux, a, b, c, d), 'o')

    elif ptype == 'both':
        _, ax1 = plt.subplots()

        ax1.set_xlabel(r'$\theta$', size=16)
        ax1.set_ylabel(r'$P(\theta)$', color='b', size=16)
        ax1.plot(thetas, p_thetas, 'b-', label=r'$P(\theta)$')
        # Make the y-axis label and tick labels match the line color.
        for tl in ax1.get_yticklabels():
            tl.set_color('b')

        ax2 = ax1.twinx()
        ax2.set_ylabel(r'$I(\theta)$', color='r', size=16)
        ax2.plot(thetas, i_thetas, 'r-', label=r'$I(\theta)$')
        for tl in ax2.get_yticklabels():
            tl.set_color('r')
        if max_info:
            aux = irt.max_info(a, b, c, d)
            plt.plot(aux, irt.inf(aux, a, b, c, d), 'o')

        if title is not None:
            ax1.set_title(title, size=18)

        ax2.annotate(
            '$a = ' + format(a) + '$\n$b = ' + format(b) + '$\n$c = ' + format(c) + '$\n$d = ' + format(d) + '$',
            bbox=dict(facecolor='white', alpha=1), xy=(.75, .05), xycoords='axes fraction')

    if filepath is not None:
        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))
        plt.savefig(filepath, bbox_inches='tight')

    if show:
        plt.show()


def gen3d_dataset_scatter(items: numpy.ndarray, title: str = None, filepath: str = None, show: bool = True):
    """Generate the item matrix tridimensional dataset scatter plot

    .. plot::

        from catsim.cat import generate_item_bank
        from catsim import plot
        items = generate_item_bank(100)
        plot.gen3d_dataset_scatter(items)

    :param items: the item matrix
    :param title: the scatter plot title
    :param filepath: the path to save the scatter plot
    :param show: whether the generated plot is to be shown
    """
    assert Axes3D
    irt.validate_item_bank(items)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(list(items[:, 0]), list(items[:, 1]), list(items[:, 2]), s=10, c='b')

    if title is not None:
        plt.title(title, size=18)

    ax.set_xlabel('a')
    ax.set_ylabel('b')
    ax.set_zlabel('c')

    if filepath is not None:
        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))
        plt.savefig(filepath, bbox_inches='tight')

    if show:
        plt.show()


def test_progress(title: str = None, simulator: Simulator = None, index: int = None, thetas: list = None,
                  administered_items: numpy.ndarray = None, true_theta: float = None, info: bool = False,
                  var: bool = False, see: bool = False, reliability: bool = False, filepath: str = None,
                  show: bool = True):
    """Generates a plot representing an examinee's test progress. Note that,
    while some functions increase or decrease monotonically, like test information
    and standard error of estimation, the plot calculates these values using the
    examinee's proficiency estimated at that given time of the test. This means
    that a test that was tought to be informative at a given point may not be as
    informative after new estimates are done.

    .. plot::

        from catsim.cat import generate_item_bank
        from catsim import plot
        from catsim.initialization import RandomInitializer
        from catsim.selection import MaxInfoSelector
        from catsim.estimation import HillClimbingEstimator
        from catsim.stopping import MaxItemStopper
        from catsim.simulation import Simulator

        initializer = RandomInitializer()
        selector = MaxInfoSelector()
        estimator = HillClimbingEstimator()
        stopper = MaxItemStopper(20)
        s = Simulator(generate_item_bank(100), 10)
        s.simulate(initializer, selector, estimator, stopper)
        plot.test_progress(simulator=s, index=0)
        plot.test_progress(simulator=s, index=0, info=True, var=True, see=True)

    :param title: the plot title.
    :param simulator: a simulator which has already simulated a series of CATs,
                      containing estimations to the examinees' proficiencies and
                      a list of administered items for each examinee.
    :param index: the index of the examinee in the simulator whose plot is to be done.
    :param thetas: if a :py:class:`Simulator` is not passed, then a list of proficiency
                   estimations can be manually passed to the function.
    :param administered_items: if a :py:class:`Simulator` is not passed, then a
                               matrix of administered items, represented by their
                               parameters, can be manually passed to the function.
    :param true_theta: the value of the examinee's true proficiency. If it is passed,
                       it will be shown on the plot, otherwise not.
    :param info: plot test information. It only works if both proficiencies and
                 administered items are passed.
    :param var: plot the estimation variance during the test. It only
                works if both proficiencies and administered items are passed.
    :param see: plot the standard error of estimation during the test. It only
               works if both proficiencies and administered items are passed.
    :param reliability: plot the test reliability. It only works if both proficiencies
                        and administered items are passed.
    :param filepath: the path to save the plot
    :param show: whether the generated plot is to be shown
    """
    if simulator is None and thetas is None and administered_items is None:
        raise ValueError('Not a single plottable object was passed.')

    plt.figure()

    if title is not None:
        plt.title(title, size=18)

    if simulator is not None and index is not None:
        thetas = simulator.estimations[index]
        administered_items = simulator.items[simulator.administered_items[index]]
        true_theta = simulator.examinees[index]

    if thetas is not None and administered_items is not None and len(thetas) - 1 != len(administered_items[:, 1]):
        raise ValueError('Number of latest_estimations and administered items is not the same. They should be.')

    # len(thetas) - 1 because the first item is made by the initializer
    xs = list(range(len(thetas))) if thetas is not None else list(range(len(administered_items[:, 1])))

    if thetas is not None:
        plt.plot(xs, thetas, label=r'$\hat{\theta}$')
    if administered_items is not None:
        difficulties = administered_items[:, 1]
        plt.plot(xs[1:], difficulties, label='Item difficulty')
    if true_theta is not None:
        plt.hlines(true_theta, 0, len(xs), label=r'$\theta$')
    if thetas is not None and administered_items is not None:

        # calculate and plot test information, var, standard error and reliability
        if info:
            infos = [irt.test_info(thetas[x], administered_items[:x + 1, ]) for x in xs]
            plt.plot(xs, infos, label=r'$I(\theta)$')

        if var:
            varss = [irt.var(thetas[x], administered_items[:x + 1, ]) for x in xs]
            plt.plot(xs, varss, label=r'$Var$')

        if see:
            sees = [irt.see(thetas[x], administered_items[:x + 1, ]) for x in xs]
            plt.plot(xs, sees, label=r'$SEE$')

        if reliability:
            reliabilities = [irt.reliability(thetas[x], administered_items[:x + 1, ]) for x in xs]
            plt.plot(xs, reliabilities, label='Reliability')
    plt.xlabel('Items')
    plt.grid()
    plt.legend(loc='best')

    if filepath is not None:
        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))
        plt.savefig(filepath, bbox_inches='tight', dpi=300)

    if show:
        plt.show()



#!/usr/bin/env python3

from setuptools import setup, find_packages
from catsim import __version__

setup(
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Education :: Testing',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)'
    ],
    name='catsim',
    version=__version__,
    description='Computerized Adaptive Testing Simulator',
    author='Douglas De Rizzo Meneghetti',
    author_email='douglasrizzom@gmail.com',
    url='https://douglasrizzo.github.io/catsim',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    install_requires=[
        'numpy', 'scipy', 'matplotlib', 'scikit-learn', 'json_tricks'
    ],
    license='GPLv2'
)



import unittest

import numpy

from catsim import cat, irt, stats, plot
from catsim.cat import generate_item_bank
from catsim.estimation import HillClimbingEstimator, DifferentialEvolutionEstimator
from catsim.initialization import RandomInitializer, FixedPointInitializer
from catsim.selection import The54321Selector, MaxInfoBBlockingSelector, MaxInfoStratificationSelector, LinearSelector, \
    AStratifiedSelector, RandomesqueSelector, AStratifiedBBlockingSelector, MaxInfoSelector, RandomSelector, \
    ClusterSelector
from catsim.simulation import Simulator
from catsim.stopping import MaxItemStopper, MinErrorStopper


def test_item_bank_generation():
    for items in [generate_item_bank(5, '1PL'), generate_item_bank(5, '2PL'), generate_item_bank(5, '3PL'),
                  generate_item_bank(5, '3PL', corr=0), generate_item_bank(5, '4PL')]:
        irt.validate_item_bank(items, raise_err=True)

    items = numpy.zeros(100)
    irt.validate_item_bank(items)
    items = irt.normalize_item_bank(items)
    irt.validate_item_bank(items, raise_err=True)


def test_plots():
    from matplotlib.pyplot import close
    initializer = RandomInitializer()
    selector = MaxInfoSelector()
    estimator = HillClimbingEstimator()
    stopper = MaxItemStopper(20)
    s = Simulator(generate_item_bank(100), 10)
    s.simulate(initializer, selector, estimator, stopper, verbose=True)

    for item in s.items[0:10]:
        yield plot.item_curve, item[0], item[1], item[2], item[3], 'Test plot', 'icc', False, None, False
        yield plot.item_curve, item[0], item[1], item[2], item[3], 'Test plot', 'iic', True, None, False
        yield plot.item_curve, item[0], item[1], item[2], item[3], 'Test plot', 'both', True, None, False
        close('all')

    plot.gen3d_dataset_scatter(items=s.items, show=False)
    plot.test_progress(title='Test progress', simulator=s, index=0, info=True, see=True, reliability=True, show=False)

    # close all plots after testing
    close('all')


def test_stats():
    import numpy.random as nprnd
    for _ in range(10):
        items = generate_item_bank(500)
        stats.coef_variation(items)
        stats.coef_correlation(items)
        stats.covariance(items)
        stats.covariance(items, False)
        stats.scatter_matrix(items)

        random_integers = nprnd.randint(30, size=1000)
        stats.bincount(random_integers)


def one_simulation(items, examinees, initializer, selector, estimator, stopper):
    s = Simulator(items, examinees)
    s.simulate(initializer, selector, estimator, stopper, verbose=True)
    cat.rmse(s.examinees, s.latest_estimations)


def test_simulations():
    examinees = 10
    test_size = 20
    bank_size = 5000

    logistic_models = [  # '1PL', '2PL',
        '4PL']

    initializers = [RandomInitializer('uniform', (-5, 5)), FixedPointInitializer(0)]
    infinite_selectors = [MaxInfoSelector(), RandomSelector()]  # , IntervalIntegrationSelector(0.3)]
    finite_selectors = [LinearSelector(list(numpy.random.randint(bank_size, size=test_size))),
                        AStratifiedSelector(test_size), AStratifiedBBlockingSelector(test_size),
                        MaxInfoStratificationSelector(test_size), MaxInfoBBlockingSelector(test_size),
                        The54321Selector(test_size), RandomesqueSelector(5)]
    estimators = [HillClimbingEstimator(), DifferentialEvolutionEstimator((-8, 8))]

    for logistic_model in logistic_models:
        for estimator in estimators:
            for initializer in initializers:
                for stopper in [MaxItemStopper(test_size)]:
                    for selector in finite_selectors:
                        items = generate_item_bank(bank_size, itemtype=logistic_model)
                        yield one_simulation, items, examinees, initializer, selector, estimator, stopper

                for stopper in [MinErrorStopper(.4), MaxItemStopper(test_size)]:
                    for selector in infinite_selectors:
                        items = generate_item_bank(bank_size, itemtype=logistic_model)
                        yield one_simulation, items, examinees, initializer, selector, estimator, stopper


def test_cism():
    from sklearn.cluster import KMeans

    examinees = 10
    initializers = [RandomInitializer('uniform', (-5, 5)), FixedPointInitializer(0)]
    estimators = [HillClimbingEstimator(), DifferentialEvolutionEstimator((-8, 8))]
    stoppers = [MaxItemStopper(20), MinErrorStopper(.4)]

    for initializer in initializers:
        for estimator in estimators:
            for stopper in stoppers:
                items = generate_item_bank(5000)
                clusters = list(KMeans(n_clusters=8).fit_predict(items))
                ClusterSelector.weighted_cluster_infos(0, items, clusters)
                ClusterSelector.avg_cluster_params(items, clusters)
                selector = ClusterSelector(clusters=clusters, r_max=.2)
                yield one_simulation, items, examinees, initializer, selector, estimator, stopper


if __name__ == '__main__':
    unittest.main()
